from meta_paper.adapters._base import PaperMetadataAdapter, PaperDetails, PaperListing
from meta_paper.adapters._open_citations import OpenCitationsAdapter
from meta_paper.adapters._semantic_scholar import SemanticScholarAdapter

__all__ = [
    "OpenCitationsAdapter",
    "PaperDetails",
    "PaperListing",
    "PaperMetadataAdapter",
    "SemanticScholarAdapter",
]



import re


class DOIPrefixMixin:
    __DOI_RE = re.compile(r"\bdoi:\b", re.IGNORECASE | re.S)

    def _prepend_doi(self, doi: str, upper_case: bool = True) -> str:
        doi = str(doi).strip()
        prefix = "DOI:" if upper_case else "doi:"
        if doi.upper().startswith("DOI:"):
            return self.__DOI_RE.sub(prefix, doi, 1)
        return f"{prefix}{doi}"

    def _has_doi_prefix(self, doi: str) -> bool:
        return bool(self.__DOI_RE.search(doi))



import asyncio
import itertools
from collections.abc import Sequence
from logging import Logger
from typing import Iterable, Generator, Callable

import httpx
from tenacity import RetryError

from meta_paper.adapters import (
    OpenCitationsAdapter,
    SemanticScholarAdapter,
    PaperListing,
    PaperDetails,
    PaperMetadataAdapter,
)
from meta_paper.logging import null_logger
from meta_paper.search import QueryParameters


class PaperMetadataClient:
    def __init__(
        self, http_client: httpx.AsyncClient | None = None, logger: Logger | None = None
    ) -> None:
        self.__providers: list[PaperMetadataAdapter] = []
        self.__http = http_client or httpx.AsyncClient(
            headers={
                "Accept": "application/json",
                "Accept-Encoding": "deflate,gzip;q=1.0",
            }
        )
        self.__logger = (logger or null_logger()).getChild(self.__class__.__name__)

    @property
    def providers(self) -> Sequence[PaperMetadataAdapter]:
        return self.__providers

    def use_open_citations(self, token: str | None = None) -> "PaperMetadataClient":
        """Add OpenCitations adapter to the client."""
        self.__providers.append(OpenCitationsAdapter(self.__http, token))
        return self

    def use_semantic_scholar(self, api_key: str | None = None):
        """Add SemanticScholar adapter to the client."""
        self.__providers.append(
            SemanticScholarAdapter(
                self.__http, api_key, self.__logger.getChild("SemanticScholarAdapter")
            )
        )
        return self

    def use_custom_provider(
        self, provider: PaperMetadataAdapter
    ) -> "PaperMetadataClient":
        self.__providers.append(provider)
        return self

    async def search(self, query: QueryParameters) -> list[PaperListing]:
        """Perform an asynchronous search across all providers."""
        tasks = [provider.search(query) for provider in self.providers]
        results = await asyncio.gather(*tasks)
        results = list(itertools.chain.from_iterable(results))
        return list(self.__dedupe_by_doi(results))

    async def get_one(self, doi: str) -> PaperDetails:
        """Fetch paper summaries asynchronously from all providers."""
        tasks = [provider.get_one(doi) for provider in self.providers]
        paper_data = []

        for coro in asyncio.as_completed(tasks):
            try:
                paper_data.append(await coro)
            except RetryError:
                self.__logger.error("retry count exceeded for doi '%s'", doi)
            except Exception as exc:
                self.__logger.fatal("generic error fetching '%s': %s", doi, exc)
                self.__logger.debug("error details", exc_info=exc)

        return self.__to_paper_details(paper_data)

    async def get_many(self, identifiers: Iterable[str]) -> Iterable[PaperDetails]:
        """Fetch paper summaries asynchronously from all providers."""
        tasks = [provider.get_many(identifiers) for provider in self.providers]
        paper_data = {}
        for coro in asyncio.as_completed(tasks):
            try:
                provider_papers = await coro
                for paper in provider_papers:
                    doi_papers = paper_data.get(paper.doi) or set()
                    doi_papers.add(paper)
                    paper_data[paper.doi] = doi_papers
            except RetryError as exc:
                self.__logger.error("retry count exceeded while fetching batch")
                self.__logger.debug("error details", exc_info=exc)
            except Exception as exc:
                self.__logger.fatal("generic error while fetching batch")
                self.__logger.debug("error details", exc_info=exc)

        return map(self.__to_paper_details, paper_data.values())

    def __to_paper_details(self, paper_data: Iterable[PaperDetails]) -> PaperDetails:
        doi = self.__longest_str(paper_data, lambda x: x.doi)
        title = self.__longest_str(paper_data, lambda x: x.title)
        abstract = self.__longest_str(paper_data, lambda x: x.abstract)
        unique_citations = list(
            sorted(set(itertools.chain.from_iterable(x.citations for x in paper_data)))
        )
        unique_references = list(
            sorted(set(itertools.chain.from_iterable(x.references for x in paper_data)))
        )
        unique_author_names = set(
            a for a in itertools.chain.from_iterable(d.authors for d in paper_data)
        )
        has_pdf = any(d.has_pdf for d in paper_data)
        pdf_url = self.__longest_str(paper_data, lambda x: x.pdf_url)
        url = self.__longest_str(paper_data, lambda x: x.url)
        source = self.__longest_str(paper_data, lambda x: x.source)
        return PaperDetails(
            doi=doi,
            title=title,
            abstract=abstract,
            source=source,
            citations=unique_citations,
            references=list(unique_references),
            authors=list(unique_author_names),
            has_pdf=has_pdf,
            pdf_url=pdf_url,
            url=url,
            year=max(d.year for d in paper_data),
        )

    @staticmethod
    def __longest_str(
        items: Iterable[PaperDetails], attr_selector: Callable[[PaperDetails], str]
    ) -> str:
        generate_attr_values = (
            attr_value for obj in items if (attr_value := attr_selector(obj))
        )
        return max(generate_attr_values, key=len, default="")

    @staticmethod
    def __dedupe_by_doi(
        results: Iterable[PaperListing],
    ) -> Generator[PaperListing, None, None]:
        """Remove duplicates based on DOI or title."""
        seen = set()
        for result in results:
            if result.doi in seen:
                continue
            seen.add(result.doi)
            yield result



import re
from typing import Iterable, Literal

import httpx
from tenacity import (
    retry,
    retry_if_exception,
    wait_exponential_jitter,
    stop_after_delay,
)

from meta_paper.adapters._base import PaperDetails, PaperListing, PaperMetadataAdapter
from meta_paper.adapters._doi_prefix import DOIPrefixMixin
from meta_paper.search import QueryParameters


def _retry_open_citations(exc: BaseException) -> bool:
    if isinstance(exc, httpx.ReadTimeout):
        return True
    if isinstance(exc, httpx.HTTPStatusError) and exc.response.status_code == 429:
        return True
    return False


class OpenCitationsAdapter(DOIPrefixMixin, PaperMetadataAdapter):
    REFERENCES_REST_API = "https://opencitations.net/index/api/v2"
    META_REST_API = "https://w3id.org/oc/meta/api/v1"
    DOI_RE = re.compile(r"^(doi:10\.\d{4,9}/\S+)$", re.IGNORECASE)

    def __init__(
        self, http_client: httpx.AsyncClient, api_token: str | None = None
    ) -> None:
        self.__http = http_client
        self.__headers = {} if not api_token else {"Authorization": api_token}

    @property
    def http_headers(self):
        return self.__headers

    async def search(self, _: QueryParameters) -> list[PaperListing]:
        return []

    async def __get_related(
        self, doi: str, relation_type: Literal["references", "citations"]
    ):
        endpoint_url = f"{self.REFERENCES_REST_API}/{relation_type}/{doi}"
        response = await self.__http.get(endpoint_url, headers=self.__headers)
        response.raise_for_status()

        citation_attr = "cited" if relation_type == "references" else "citing"
        return [
            self.DOI_RE.search(ref[citation_attr]).group(1)
            for ref in response.json()
            if self.DOI_RE.search(ref[citation_attr])
        ]

    @retry(
        retry=retry_if_exception(_retry_open_citations),
        wait=wait_exponential_jitter(max=10),
        stop=stop_after_delay(10),
    )
    async def get_one(self, doi: str | Iterable[str]) -> PaperDetails:
        """Fetch references and citations for a DOI."""
        doi = self._prepend_doi(doi, False)
        if not self.DOI_RE.match(doi):
            raise ValueError(f"{doi} is not a valid DOI")

        refs = await self.__get_related(doi, "references")
        citations = await self.__get_related(doi, "citations")

        response = await self.__http.get(
            f"{self.META_REST_API}/metadata/{doi}", headers=self.__headers
        )
        response.raise_for_status()
        metadata = next(iter(response.json()))
        pub_date_parts = metadata.get("pub_date", "").split("-")
        year = (
            int(pub_date_parts[0])
            if len(pub_date_parts) > 0 and pub_date_parts[0]
            else 0
        )

        return PaperDetails(
            doi=doi,
            title=metadata["title"],
            authors=metadata.get("authors", "").split(";"),
            abstract="",
            references=refs,
            citations=citations,
            source=metadata.get("venue", ""),
            url=f"https://dx.doi.org/{doi}",
            year=year,
        )

    async def get_many(self, identifiers: Iterable[str]) -> Iterable[PaperDetails]:
        return []
